/**************************************************************************************************
  SD_Driver.h



  Copyright (C) 2007  Val Gretchev

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************************************/
#ifndef SD_DRIVER_H
#define SD_DRIVER_H


#include "__Lib_FAT32.h"



// Defines
#define SD_SPI_CR1 0x0000035C
#define SD_SW_DELAY 50
#define SD_DELAY_CNTR 1
#define USB_DELAY 50

// Message constants
#define MSG_DATA_LENGTH  2   // Message data length
#define MSG_QUEUE_LENGTH 10  // Message queue length



// Diagnostic output for SD_Driver ****************************************************************
#define SD_DEBUG 1     // Set to 0 to prevent diagnostic messages from SD_Driver

#if SD_DEBUG == 1
//#define SD_DEBUGOUT(c) UART2_PutQueue(c)
#define SD_DEBUGOUT(c) USB_PutQueue(c)
#else
#define SD_DEBUGOUT(c) {}
#endif



// Typedefs
//Types of messages that can be sent to the SD card task.
typedef enum {
        MSG_GETDIR,
        MSG_GETFILE,
        MSG_WRITELINE
}
TSDMsgTypes;

// Structure of a message that is sent to the SD card task.
typedef struct
{
     TSDMsgTypes msgType;               // Message type
     uint8_t msgData[MSG_DATA_LENGTH];  // Message data
}
TSDcardMessage;


// Function Prototypes
void SD_Card_Detect_Enable(void);
void SD_InitIO(void);
void SD_Card_Initialize(void);
void SPI3_Set_SD_Mode_Slow(void);
void SPI3_Set_SD_Mode_Fast(void);
short InitFat32(void);

// Common file handling functions
//short SetGlobalTime(__TIME *pTM);
short SetGlobalTime(void);
short FormatSD(char *VolumeLabel);
short ChangeDir(char *DirName);
short MakeDir(char *DirName);
short CreateFile(unsigned short num, char *filename);
short CloseFile(unsigned short num);
short OpenFile(unsigned short, char *filename);
short DeleteFile(char *name);
short CheckEOF(unsigned short num);

// Text file handling functions
short WriteTextLine(unsigned short num, char *buf);
short ReadTextLine(unsigned short, char *buf);

// File services functions using state machine
short GetDir(void);
char GetFile(char *name);
char WriteLine(char *name);

// File services state machine
void DoFileSM(void);

// Binary file handling functions
short BinToAscii(char *InputBuf, char *OutputBuf, char ByteCount);
short AsciiToBin(char *InputBuf, char *OutputBuf, char ByteCount);

short WriteBinLine(unsigned short num, char *buf);
short ReadBinLine(unsigned short num, char *buf);
void ReadWholeBinFile(char *name);

void TestBinToAsc(void);
void TestAscToBin(void);

// Test functions
void WriteBinFile(void);
void ReadBinFile(void);


// Diagnostics
void ShowSPI3Regs(void);
void CheckStateOfCardDetect(void);

// Additional helper functions
short StripControlChar(char* StrIn, char* StrOut, int Count);
void StripCtrlCharExceptTab(char* StrIn);
unsigned long CRC32Calc(unsigned long vCRC32, char *Buff, int Size);
//unsigned long CRC32Calc(unsigned long vCRC32, char *Buff);


void ShowErrorCode(void);
void SDcardDetect(void);

#endif

/**************************************************************************************************
  End of File
**************************************************************************************************/