/**************************************************************************************************
  WiFi_ESP_Driver.c

  Copyright (C) 2007  Val Gretchev

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************************************
  WiFi ESP click should be placed in slot 1.


**************************************************************************************************/


// Include Files
#include <stdint.h>
#include <stddef.h>
#include <built_in.h>
#include "FreeRTOSConfig.h"
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"
#include "timers.h"
#include "projdefs.h"
#include "portmacro.h"

#include "Common.h"
#include "WiFi_ESP_Driver.h"
#include "USART3_driver.h"
#include "Buffers.h"
#include "USB_Driver3.h"
#include "USART2_driver.h"
#include "__Time.h"
#include "RTC.h"


// Shield slot 1 IO pins
sbit WiFiESP_EN  at GPIOD_ODR.B13;  // Slot 1 CS
sbit WiFiESP_RST at GPIOC_ODR.B2;   // Slot 1 RST

// Constants
const char SSID[] = "Rigel";                // Put the SSID name of you WiFi router here
const char Password[] = "secret";           // Use your router password here





// Headers for web page
char httpHeader1[] = "HTTP/1.1 200 OK\r\n";
char httpHeader2[] = "Content-type: text/html\r\n";
char httpHeader3[] = "\r\n\r\n";

// Simple web page that can be customized
char htmlPage[] = "<html>\
<!-- Text between angle brackets is an HTML tag and is not displayed. \
Most tags, such as the HTML and /HTML tags that surround the contents of \
a page, come in pairs; some tags, like HR, for a horizontal rule, stand \
alone. Comments, such as the text you're reading, are not displayed when \
the Web page is shown. The information between the HEAD and /HEAD tags is \
not displayed. The information between the BODY and /BODY tags is displayed.-->\
<head>\
<title>Enter a title, displayed at the top of the window.</title>\
</head>\
<!-- The information between the BODY and /BODY tags is displayed.-->\
<body>\
<h1>Enter the main heading, usually the same as the title.</h1>\
<p>Be <b>bold</b> in stating your key points. Put them in a list: </p>\
<ul>\
<li>The first item in your list</li>\
<li>The second item; <i>italicize</i> key words</li>\
</ul>\
<p>Improve your image by including an image. </p>\
<hr>\
<!-- And add a copyright notice.-->\
<p>&#169; Wiley Publishing, 2011</p>\
</body>\
</html>";


// Variables
char WiFiESP_Status;

// www.google.com         8.8.8.8
// www.rogers.com         64.71.255.204
// resolver1.opendns.com  208.67.222.222 https://dnslytics.com/ip/208.67.222.222
unsigned char httpMethod[] = "GET /";
char DNS_SErverIP[18]   = {"208.67.222.222"};    // Using this DNS server

// This Time server returns 3 IP addresses: 132.246.11.238 132.246.11.227 132.246.11.229
char NTP_ServerName[] = {"time.nrc.ca"};
char NTP_ServerIP[16] = {"132.246.11.238"}; // Using this NTP server

// This address has to be changed manually in order to communicate with applications on the PC
char PC_IP_Address[16] = {"192.168.0.14"};  // My PC address assigned by DHCP of the router
//char PC_IP_Address[16] = {"192.168.4.2"};


// This address will be set when WiFi ESP connects to the router
char Local_IP_Addr[16];                     // WiFi ESP address assigned by DHCP of the router

// Buffers
char QueryBuffer[256];                      // The DNS query is assembled in this buffer
unsigned  QueryBufLen;                      // Holds the length of the query message
char IP_Address[4][16];                     // Buffers to hold 4 IP addresses
char DNS_Answer[256];                       // Buffer to store the DNS answer
char NTP_Query[48];                         // The NTP query is assembled in this buffer
char NTP_Answer[SIZE_OF_NTP_BUF];           // Buffer to store the NTP answer
char UDP_Buffer[128];                       // Used in DNS request
char TCP_Client[8];                         // The TCP server can support 4 clients

unsigned long epoch;


// External variables
extern TTime MyTime;
extern char Transparent;
extern char U3RcvBuf[U3_RCV_BUF_SIZE];
extern char U3RcvBufPtr;
extern const code char indexPage1[];
extern char U3BigBuf[U3_XMT_BUF_SIZE];





// Buffers
char Messages[NUM_MESSAGES][MESSAGE_SIZE];  // Messages from the ESP chip are saved in this buffer
short MsgCntr = -1;                         // Points to the last message buffer used

// Semaphores
SemaphoreHandle_t vWiFiESPSemHandle;        // This semaphore is not used for now

// Queues
QueueHandle_t vWiFiESPPointerQueue;         // Queue of pointers
QueueHandle_t CmdQueue;                     // Command queue

// Structures
TimeStruct my_time;                         // Used when obtaining time from the NTP server


/**************************************************************************************************
  Test an IP address for validity.
  Inputs: pointer to IP address buffer
  Returns: -1 if something is wrong with the address (see message displayed), 0 if valid IP
**************************************************************************************************/
short CheckForValidIP(char *Addr) {
      char *BufNum;
      char  Copy[16];
      char  Test[8];
      short Stat;
      int   Number, i, j;

      if (strlen(Addr) == 0) {         // String is empty test
          USB_PutQueue("Error: Input string is empty\r\n");
          return -1;
      }

      if (strlen(Addr) > 15) {         // String is too long test
          USB_PutQueue("Error: Input string is too long\r\n");
          return -1;
      }
#if WIFI_DEBUG
      BufNum = GetBuffer();            // Diagnostic output
      sprintf(BufNum,"Diag: Size of input string: %d\r\n",strlen(Addr));
      USB_PutQueue(BufNum);
#endif

      memset(Copy,0,sizeof(Copy));     // Zero out local buffer
      strcpy(Copy,Addr);               // Copy the IP address to local buffer

#if WIFI_DEBUG
      BufNum = GetBuffer();            // Diagnostic output
      sprintf(BufNum,"Diag: Size of string in copy: %d\r\n",strlen(Copy));
      USB_PutQueue(BufNum);
#endif

#if WIFI_DEBUG
      BufNum = GetBuffer();            // Diagnostic output
      strcpy(BufNum,Copy);
      USB_PutQueue(BufNum);
#endif

      for (j=0;j<3;j++) {
          Stat = GetToken(Test,Copy,'.');   // First 3 fields are processed similarly
#if WIFI_DEBUG
          BufNum = GetBuffer();        // Diagnostic output
          sprintf(BufNum,"Diag: Processing number %d: %s\r\n",j+1,Test);
          USB_PutQueue(BufNum);
#endif
          if (Stat == 0) {             // IP address not enough fields test
              USB_PutQueue("Error: IP string has too few numbers\r\n");
              return -1;
          }

          if ((strlen(Test) == 0) || (strlen(Test) > 3)) { // Number of characters in field test
              USB_PutQueue("Error: A number needs to have 1 to 3 digits\r\n");
              return -1;
          }

          for (i=0;i<strlen(Test);i++) {    // Numeric character test
              if ((Test[i] >= 48) && (Test[i] <= 57)) {

              } else {                      // Error output
                   USB_PutQueue("Error: The number contains a non-numeric character\r\n");
                   return -1;
              }
          }

          if ((strlen(Test) > 1) && (Test[0] == '0')) {    // Leading zero test
              USB_PutQueue("Error: The number has leading zero\r\n");
                 return -1;
          }

          Number = atoi(Test);
          if (Number > 255) {               // Field range test
              USB_PutQueue("Error: The number is greater than 255\r\n");
              return -1;
          }
      }

      Stat = GetToken(Test,Copy,'.');       // Fourth field should not have more after it
#if WIFI_DEBUG
      BufNum = GetBuffer();                 // Diagnostic output
      sprintf(BufNum,"Diag: Processing number 4: %s\r\n",Test);
      USB_PutQueue(BufNum);
#endif
      if (Stat != 0) {                      // Too many fields
          USB_PutQueue("Error: IP string has too many numbers\r\n");
          return -1;
      }

      if ((strlen(Test) == 0) || (strlen(Test) > 3)) {     // A field must have 1 to 3 digits
          USB_PutQueue("Error: A number needs to have 1 to 3 digits\r\n");
          return -1;
      }

      for (i=0;i<strlen(Test);i++) {        // Non-numeric character test
           if ((Test[i] >= 48) && (Test[i] <= 57)) {

           } else {
               USB_PutQueue("Error: The number contains a non-numeric character\r\n");
               return -1;
           }
      }

      if ((strlen(Test) > 1) && (Test[0] == '0')) {   // Test for leading zero
          USB_PutQueue("Error: The number has leading zero\r\n");
               return -1;
      }

      Number = atoi(Test);
      if (Number >255) {                    // Range test
          USB_PutQueue("Error: The number is greater than 255\r\n");
          return -1;
      }

      return 0;
}

/**************************************************************************************************
  Utility functions.
**************************************************************************************************/
/**************************************************************************************************
  Given a string s that contains tokens delimited by a character c, this function strips the first
  token from the string and puts it in t. The return code is -1 if the string is empty, 0 if there
  are no more tokens following this one, 1 if there are more tokens to process. String s is
  destroyed in the process.
**************************************************************************************************/
short GetToken(char *t, char *s, char c) {
     char *p;                       // Token pointer

     if (strlen(s) == 0) {          // If the string is empty
         return (-1);
     }
     p = strchr(s,c);               // Find location of the test character
     if (p == NULL) {               // If didn't find the test characters
         strcpy(t,s);               // Copy the final token
         strcpy(s,"");              // Nullify the input string
         return 0;
     }
     s[strlen(s)-strlen(p)] = '\0'; // Replace the text character with a null
     strcpy(t,s);                   // Copy the token
     p++;                           // Advance the pointer to the next character position
     strcpy(s,p);                   // Overlay the input string minus the first token
     return 1;
}

/**************************************************************************************************
  Copy from block to dest for size number of bytes.
**************************************************************************************************/
void ByteCopy(unsigned char *block, unsigned char *dest, int size) {
    int i = 0;

    while (i < size){
        dest[i] = block[i];
        i++;
    }
    return;
}

/**************************************************************************************************
  Print size number of characters from array with blank between bytes.
**************************************************************************************************/
int HexPrint( unsigned char *array, int size) {
    char *BufNum;
    char TempBuf[4];
    int i;

    if (size < 1) {
        USB_PutQueue("Size is 0\r\n");
        return -1;
    }

    //USB_PutQueue("Hex string: \r\n");
    USB_PutQueue("Hex string: ");
    BufNum = GetBuffer();

    for(i=0;i<size;i++) {
        sprintf(TempBuf,"%02X ",array[i]);
        strcat(BufNum,TempBuf);
    }
    strcat(BufNum,"\r\n");
    USB_PutQueue(BufNum);
    return 1;
}

/**************************************************************************************************
  WiFiESP - Initialize.
**************************************************************************************************/
void WiFiESP_Initialize(void) {

     WiFiESP_EN = 0;   // Set outputs to 0 before enabling GPIO
     WiFiESP_RST = 0;
     GPIO_Digital_Output (&GPIOD_BASE, _GPIO_PINMASK_13);    // EN
     GPIO_Digital_Output (&GPIOC_BASE, _GPIO_PINMASK_2);     // RST

     WIFI_DEBUGOUT("Diag: WiFiESP IO assigned\r\n");
}

/**************************************************************************************************
  WiFiESP - Hardware Reset.
**************************************************************************************************/
void WiFiESP_HW_Reset(void) {

     WiFiESP_EN = 1;
     vTaskDelay(pdMS_TO_TICKS(500));
     WiFiESP_RST = 1;
     vTaskDelay(pdMS_TO_TICKS(500));

     WIFI_DEBUGOUT("Diag: Resetting WiFiESP hardware\r\n");
}

/**************************************************************************************************
  WiFiESP - Software Reset.
**************************************************************************************************/
void WiFiESP_SW_Reset(void) {

     WIFI_DEBUGOUT("Diag: Sending software reset command\r\n");
     USART3_PutQueueText("AT+RST\r\n");
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  WiFiESP - Restore the Factory Default Settings.
**************************************************************************************************/
void WiFiESP_Set_Factory_Defaults(void) {

     WIFI_DEBUGOUT("Diag: Restoring factory defaults\r\n");
     USART3_PutQueueText("AT+RESTORE\r\n");
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  WiFiESP - Set parameters in flash.
**************************************************************************************************/
void WiFiESP_Configure_Flash(void) {
     char *BufNum;
     
     WIFI_DEBUGOUT("Diag: Setting WiFi mode\r\n");
     USART3_PutQueueText("AT+CWMODE_DEF=3\r\n");      // SoftAP and Station mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Connecting to router\r\n");
     BufNum = GetBuffer();
     sprintf(BufNum,"AT+CWJAP_DEF=\"");
     strcat(BufNum,(char*)SSID);                      // My home router SSID
     strcat(BufNum,"\",\"");
     strcat(BufNum,(char*)Password);                  // My home router password
     strcat(BufNum,"\"\r\n");
     USART3_PutQueueText(BufNum);
     LookForPrompt(portMAX_DELAY,"OK",1);
     
     // This allows WiFiESP to act as a router. The SSID below should be visible in the PC's list.
     WIFI_DEBUGOUT("Diag: Configuring SoftAP\r\n");
     USART3_PutQueueText("AT+CWSAP_DEF=\"WiFiESP\",\"ClickBoard\",5,3,1,0\r\n"); // SSID, password
     LookForPrompt(portMAX_DELAY,"OK",1);

     //
     WIFI_DEBUGOUT("Diag: Setting SoftAP and Station mode, enabling DHCP\r\n");
     USART3_PutQueueText("AT+CWDHCP_DEF=2,1\r\n");    // SoftAP and Station, DHCP enabled
     LookForPrompt(portMAX_DELAY,"OK",1);
     //
     
     // This sets the range of IP addresses that SoftAP will issue to connections.
     WIFI_DEBUGOUT("Diag: Setting range of IP addresses for SoftAP DHCP\r\n");
     USART3_PutQueueText("AT+CWDHCPS_DEF=1,3,\"192.168.4.10\",\"192.168.4.15\"\r\n");
     LookForPrompt(portMAX_DELAY,"OK",1);
     
     WiFiESP_Set_UART_Config();
}

/**************************************************************************************************
  WiFiESP - Save UART Configuration in Flash. Factory default is 115,200 baud. If serial
            interface is dropping characters, change to 9600 baud using this function.
            Actually, the command "AT+UART_CUR?\r\n" returns +UART_CUR:115273,8,1,0,1
            Send 'e' from USB Utility to set UART to the values in the command below.
**************************************************************************************************/
void WiFiESP_Set_UART_Config(void) {

     WIFI_DEBUGOUT("Diag: Setting default UART configuration in flash\r\n");
     // 9600 baud, 8-bits, 1 stop bit, no parity, no hardware flow control
     //USART3_PutQueueText("AT+UART_DEF=9600,8,1,0,0\r\n");
     // 115200 baud, 8-bits, 2 stop bits, no parity, no hardware flow control
     USART3_PutQueueText("AT+UART_DEF=115200,8,2,0,0\r\n");
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  WiFiESP - Send AT command as a test.
**************************************************************************************************/
void SendAT(void) {

     WIFI_DEBUGOUT("Diag: Sending AT command\r\n");
     USART3_PutQueueText("AT\r\n");              // Test
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  WiFiESP - List available APs (access points).
**************************************************************************************************/
void ListAvailAPs(void) {

     WIFI_DEBUGOUT("Diag: The following APs were found\r\n");
     USART3_PutQueueText("AT+CWLAP\r\n");
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  WiFiESP - Wait For Response. Only used when USART3 in normal ASCII mode not in transparent.
            Returns 1 if message received, 0 if timeout
**************************************************************************************************/
char WaitForResponse(unsigned long Timeout) {
     char *pcReceivedAddr;    // Last pointer to message obtained from pointer queue
     BaseType_t rtnCode;      // Function return code
     char *BufNum;

     rtnCode = xQueueReceive(vWiFiESPPointerQueue, &pcReceivedAddr, Timeout); // Block here
     if (rtnCode == pdPASS) {
         USB_PutQueue(pcReceivedAddr); // Display the message that was received
         // This buffer saves Messages (up to 12) for processing later
         // When MsgCntr reaches 12, all further responses will be saved in 12
         if (MsgCntr < 18) {
             MsgCntr++;
             strcpy(Messages[MsgCntr],pcReceivedAddr); // Save the message for later processing

          }

         /* Enable if required
         // Diagnostic to see how many messages have been saved
         BufNum = GetBuffer();
         sprintf(BufNum,"Diag: Message counter is at %d\r\n",MsgCntr);
         USB_PutQueue(BufNum);
         */

         return 1;      // Message received
     } else {
         return 0;      // Timeout
     }
}

/**************************************************************************************************
  WiFiESP  - LookForPrompt. Only used when USART3 in Normal ASCII mode.
  Inputs:  - Timeout is passed on to WaitForResponse in milliseconds.
           - Prompt is a pointer to the expected response.
           - RstMsgCntr asks to reset MsgCntr if 1, leave it alone if 0
  Outputs: - Returns 1 if returned message compares, 0 if timeout
**************************************************************************************************/
short LookForPrompt(unsigned long Timeout, char *Prompt, char RstMsgCntr) {
     //short RetryCnt = 0; // Not used
     char  RtnVal;
     char *BufNum;

     if (RstMsgCntr) MsgCntr = -1;     // Reset message counter if requested to do so

     do {
         RtnVal = WaitForResponse(Timeout);
         //
         if (RtnVal == 0) {            // Break the loop if timeout occurres
             break;
         }
         //
         
         //RetryCnt += 1; // Not used
     } while (( strstr(Messages[MsgCntr],Prompt)) == 0); // Break the loop when Prompt arrives

     /* Diagnostic. Enable if required.
     BufNum = GetBuffer();
     sprintf(BufNum,"Timeout: %08LX\r\n",Timeout);
     USB_PutQueue(BufNum);
     USB_PutQueue(Prompt);
     BufNum = GetBuffer();
     sprintf(BufNum,"RstMsgCntr: %d\r\n",RstMsgCntr);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"Diag: Retry count: %d\r\n",RetryCnt);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"Diag: Value of message counter: %d\r\n",MsgCntr);
     USB_PutQueue(BufNum);
     */

     return (RtnVal);
}

/**************************************************************************************************
  WiFiESP - Wait For IP initial ready.
**************************************************************************************************/
void WaitForReady(void) {

     WIFI_DEBUGOUT("Diag: Waiting for initial ready\r\n");
     LookForPrompt(3000,"ready",1);
}

/**************************************************************************************************
  WiFiESP - Wait For IP Confirmation. If the WiFiESP module has not been initialized, the
            LookForPrompt function returns timeout. Send 'd' command from USB Utility to
            program the flash in WiFiESP module with your SSID and password.
**************************************************************************************************/
void WaitForIP(void) {
     short RtnVal;
     
     WaitForReady();
     
     WIFI_DEBUGOUT("Diag: Waiting for IP\r\n");

     //LookForPrompt(portMAX_DELAY,"GOT IP",1);
     RtnVal = LookForPrompt(3000,"GOT IP",1);   // Wait 3 seconds

     if (RtnVal != 0) {                         // No timeout on GOT IP
         WIFI_DEBUGOUT("Diag: Suppressing echo\r\n");
         USART3_PutQueueText("ATE0\r\n");        // Suppress echo
         LookForPrompt(portMAX_DELAY,"OK",1);

         GetLocalIP();                           // DHCP assigned IP address for the WiFi ESP

         USB_PutQueue("Diag: Ready to process commands\r\n");
     } else {                                    // Timeout on GOT IP
         USB_PutQueue("Diag: Need to configure WiFiESP module with SSID and password\r\n");
     }

}

/**************************************************************************************************
  WiFiESP - Get IP address assigned by DHCP. Display the local IP address on USB.
**************************************************************************************************/
void GetLocalIP(void) {
     short RtnVal, Return;
     char Temp[20];
     char *BufNum;

     WIFI_DEBUGOUT("Diag: Getting my IP address\r\n");
     USART3_PutQueueText("AT+CIFSR\r\n");        // Use after connecting to AP

     RtnVal = LookForPrompt(5000,"OK",1);

     memset(Local_IP_Addr,0,sizeof(Local_IP_Addr));        // Zero out local IP address location
     Return = GetToken(Temp,Messages[MsgCntr-2],'"');      // IP address follows this token
     if (Return > 0) {
         memset(Temp,0,sizeof(Temp));                      // Prepare buffer to receive address
         GetToken(Temp,Messages[MsgCntr-2],'"');           // Should be in this buffer
         strcpy(Local_IP_Addr,Temp);                       // Put it in address location
     }
     USB_PutQueue("My IP address is below:\r\n");          // Should always display
     BufNum = GetBuffer();
     strcpy(BufNum,Local_IP_Addr);                         // May change from time to time
     strcat(BufNum,"\r\n");
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  WiFiESP - Send a TCP packet. Create a TCP client port, connect to a remote server and send a
            test message. Wait for a response and then close the TCP port.
**************************************************************************************************/
void ClientTCP(void) {
     char *BufNum;

     if (CheckForValidIP(PC_IP_Address) == 0) {  // Make sure we have a valid IP address
         WIFI_DEBUGOUT("Diag: The IP is valid\r\n");
     } else {
         USB_PutQueue("Error: The IP is not valid\r\n");
         return;
     }

     WIFI_DEBUGOUT("Diag: Creating TCP client\r\n");
     // Connection to "Remote IP", "Remote port"
     //USART3_PutQueueText("AT+CIPSTART=\"TCP\",\"192.168.0.14\",1502\r\n");
     BufNum = GetBuffer();
     sprintf(BufNum,"AT+CIPSTART=\"TCP\",\"");   // Start assembling AT command
     strcat(BufNum,PC_IP_Address);               // Insert my PC IP address
     strcat(BufNum,"\",1502\r\n");               // Add in remote port
     USART3_PutQueueText(BufNum);
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Setting data length\r\n");
     USART3_PutQueueText("AT+CIPSEND=10\r\n");   // Set data length in bytes
     LookForPrompt(portMAX_DELAY,"OK",1);

     LookForPrompt(portMAX_DELAY,">",1);

     WIFI_DEBUGOUT("Diag: Sending data\r\n");
     WIFI_DEBUGOUT("Data: 001,test\r\n");
     USART3_PutQueueText("001,test\r\n");        // Send data
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Waiting for acknowledgement\r\n");
     LookForPrompt(portMAX_DELAY,"ACK,001",1);

     WIFI_DEBUGOUT("Diag: Closing the TCP connection\r\n");
     USART3_PutQueueText("AT+CIPCLOSE\r\n");    // Close the TCP connection
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  WiFiESP - Helper function. Creates an AT+CIPSEND message with the correct length of the array
            passed to it with a pointer to Msg.
**************************************************************************************************/
short SendLength(char *Msg) {
     char *BufNum;
     char LengthVal[8];
     int  Length;

     Length = strlen(Msg);                       // Get the length of the string
     sprintf(LengthVal,"%d",Length);             // Convert it to an ASCII string
     BufNum = GetBuffer();                       // Allocate a buffer
     sprintf(BufNum,"AT+CIPSEND=0,");            // Put the command into the buffer
     strcat(BufNum,LengthVal);                   // Append the length string
     strcat(BufNum,"\r\n");                      // Append return, new line
     USART3_PutQueueText(BufNum);                // Send the AT command
     //USART2_PutQueue(BufNum);                    // Test

     return 0;
}

/**************************************************************************************************
  WiFiESP - Helper function. The WiFiESP module can handle messages of only 2048 bytes. When
            sending a long text message such as a web page, this function breaks them up and
            sends the long text message in pieces. If the message is in code space (flash),
            the compiler issues a "suspicious pointer conversion" message. The (char*) typecast
            gets rid of that message so that real messages you want to know about are not hidden.
**************************************************************************************************/

short SendLongMessage(const code char *Msg) {
     char *BufNum;
     char LengthVal[8];
     int  Length, Index, i;

     Index = 0;
     Length = strlen((char*)Msg);                // Get the length of the transmit message
     if (Length > 2040) {
         while (Length > 2040) {
             sprintf(LengthVal,"%d",2040);       // Convert max length to an ASCII string
             BufNum = GetBuffer();               // Allocate a buffer
             sprintf(BufNum,"AT+CIPSEND=0,");    // Put the command into the buffer
             strcat(BufNum,LengthVal);           // Append the length string
             strcat(BufNum,"\r\n");              // Append return, new line

             WIFI_DEBUGOUT("Diag: Sending data length\r\n");
             USART3_PutQueueText(BufNum);        // Send the AT command
             // Skip the "OK" prompt
             LookForPrompt(portMAX_DELAY,">",1);

             memset(U3BigBuf,0,sizeof(U3BigBuf));
             for (i=0;i<2040;i++) {
                 U3BigBuf[i] = *Msg++;
             }
             WIFI_DEBUGOUT("Diag: Sending data\r\n");
             USART3_PutQueueText(U3BigBuf);
             LookForPrompt(portMAX_DELAY,"SEND OK",1);

             Length -= 2040;
         }

         if (Length > 0) {
             sprintf(LengthVal,"%d",Length);     // Convert remaining length to an ASCII string
             BufNum = GetBuffer();               // Allocate a buffer
             sprintf(BufNum,"AT+CIPSEND=0,");    // Put the command into the buffer
             strcat(BufNum,LengthVal);           // Append the length string
             strcat(BufNum,"\r\n");              // Append return, new line

             WIFI_DEBUGOUT("Diag: Sending data length\r\n");
             USART3_PutQueueText(BufNum);        // Send the AT command
             // Skip the "OK" prompt
             LookForPrompt(portMAX_DELAY,">",1);

             memset(U3BigBuf,0,sizeof(U3BigBuf));
             for (i=0;i<Length;i++) {
                 U3BigBuf[i] = *Msg++;
             }
             WIFI_DEBUGOUT("Diag: Sending data\r\n");
             USART3_PutQueueText(U3BigBuf);
             LookForPrompt(portMAX_DELAY,"SEND OK",1);

             Length = 0;
         }

     } else {
         if (Length > 0) {
             sprintf(LengthVal,"%d",Length);     // Convert remaining length to an ASCII string
             BufNum = GetBuffer();               // Allocate a buffer
             sprintf(BufNum,"AT+CIPSEND=0,");    // Put the command into the buffer
             strcat(BufNum,LengthVal);           // Append the length string
             strcat(BufNum,"\r\n");              // Append return, new line

             WIFI_DEBUGOUT("Diag: Sending data length\r\n");
             USART3_PutQueueText(BufNum);        // Send the AT command
             // Skip the "OK" prompt
             LookForPrompt(portMAX_DELAY,">",1);

             memset(U3BigBuf,0,sizeof(U3BigBuf));
             for (i=0;i<Length;i++) {
                 U3BigBuf[i] = *Msg++;
             }
             WIFI_DEBUGOUT("Diag: Sending data\r\n");
             USART3_PutQueueText(U3BigBuf);
             LookForPrompt(portMAX_DELAY,"SEND OK",1);

             Length = 0;
         }
     }
     WIFI_DEBUGOUT("Diag: End of data\r\n");
     return 0;
}


/**************************************************************************************************
  WiFiESP - Open a TCP port for HTTP server.
**************************************************************************************************/
void ServerTCPforHTTP(void) {
     char *BufNum;
     char LengthVal[8];
     int  Length;

     WIFI_DEBUGOUT("Diag: Disabling multiple connections\r\n");
     USART3_PutQueueText("AT+CIPMUX=0\r\n");
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Setting WiFi mode\r\n");
     USART3_PutQueueText("AT+CWMODE=3\r\n"); // SoftAP+station mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Setting normal transmission mode\r\n");
     USART3_PutQueueText("AT+CIPMODE=0\r\n"); // Must be 0 for multiple connections
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Enabling multiple connections\r\n");
     USART3_PutQueueText("AT+CIPMUX=1\r\n"); // Mode = 1: multiple connections
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Creating TCP server\r\n");
     USART3_PutQueueText("AT+CIPSERVER=1,80\r\n"); // Mode = 1: creates server, HTTP server port
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Waiting for TCP connect\r\n");
     LookForPrompt(portMAX_DELAY,"CONNECT",1);
     // A TCP port can be connected to 4 clients
     GetToken(TCP_Client,Messages[MsgCntr],',');
#if WIFI_DEBUG
     BufNum = GetBuffer();
     sprintf(BufNum,"Info: Client connected: %s\r\n",TCP_Client); // Display the TCP client number
     USB_PutQueue(BufNum);
#endif
     WIFI_DEBUGOUT("Diag: Waiting for HTTP request\r\n");
     //LookForPrompt(portMAX_DELAY,"OK",1);
     LookForPrompt(portMAX_DELAY,"Alive",1);

     if (memcmp(Messages[1],httpMethod,5)) {
         WIFI_DEBUGOUT("Info: Got GET request\r\n");
     } else {
         WIFI_DEBUGOUT("Diag: Closing the server\r\n");
         USART3_PutQueueText("AT+CIPSERVER=0\r\n");  // Close the server
         LookForPrompt(portMAX_DELAY,"OK",1);
         return;
     }

     // Use helper function to compose the required AT command
     WIFI_DEBUGOUT("Diag: Sending header 1 length\r\n");
     SendLength(httpHeader1);                    // Set AT command with data length in bytes
     // Skip the "OK" prompt
     LookForPrompt(portMAX_DELAY,">",1);


     WIFI_DEBUGOUT("Diag: Sending header 1\r\n");
     USART3_PutQueueText(httpHeader1);           // Sending the first header
     LookForPrompt(portMAX_DELAY,"SEND OK",1);

     WIFI_DEBUGOUT("Diag: Sending header 2 length\r\n");
     SendLength(httpHeader2);                    // Set AT command with data length in bytes
     // Skip the "OK" prompt
     LookForPrompt(portMAX_DELAY,">",1);

     WIFI_DEBUGOUT("Diag: Sending header 2\r\n");
     USART3_PutQueueText(httpHeader2);           // Sending the second header
     LookForPrompt(portMAX_DELAY,"SEND OK",1);

     WIFI_DEBUGOUT("Diag: Sending header 3 length\r\n");
     SendLength(httpHeader3);                    // Set AT command with data length in bytes
     // Skip the "OK" prompt
     LookForPrompt(portMAX_DELAY,">",1);

     WIFI_DEBUGOUT("Diag: Sending header 3\r\n");
     USART3_PutQueueText(httpHeader3);           // Sending the third header
     LookForPrompt(portMAX_DELAY,"SEND OK",1);

/*************************************************************************************************/

     vTaskDelay(pdMS_TO_TICKS(4000));
     USB_PutQueue("Diag: Sending web page\r\n");
     
     // Chose the web page you want to send
     //SendLongMessage((char*)indexPage1);         // Send data for long web page
     SendLongMessage(htmlPage);                  // Web page you can customize

/*************************************************************************************************/
     //
     WIFI_DEBUGOUT("Diag: Closing the server\r\n");
     USART3_PutQueueText("AT+CIPSERVER=0\r\n");  // Close the server
     LookForPrompt(portMAX_DELAY,"OK",1);
     //
}

/**************************************************************************************************
  Test for HTTP transmission. The Messages come out on the connector labeled USB UART B.
  A serial port application must be set up to communicate with this COM port. Sending command 'b'
  from USART2_Driver will execute this function (see USART2_driver.c).
**************************************************************************************************/
void TestForHTTP(void) {
     char *BufNum;
     char AT_CmdBuf[20];
     char LengthVal[8];
     int  Length;

     USART2_PutQueue("AT+CWMODE=3\r\n");         // softAP+station mode
     USART2_PutQueue("AT+CIPMODE=0\r\n");        // Must be 0 for multiple connections
     USART2_PutQueue("AT+CIPMUX=1\r\n");         // Mode = 1: multiple connections
     USART2_PutQueue("AT+CIPSEND=0,17\r\n");     // Set data length in bytes
     USART2_PutQueue(httpHeader1);               // Send data
     USART2_PutQueue("AT+CIPSEND=0,25\r\n");     // Set data length in bytes
     USART2_PutQueue(httpHeader2);               // Send data
     USART2_PutQueue("AT+CIPSEND=0,4\r\n");      // Set data length in bytes
     USART2_PutQueue(httpHeader3);               // Send data

     Length = strlen(htmlPage);                  // Very long message
     sprintf(LengthVal,"%d",Length);             // Convert length of message
     BufNum = GetBuffer();                       // Get a free buffer
     sprintf(BufNum,"AT+CIPSEND=0,");            // Start building AT message in buffer
     strcat(BufNum,LengthVal);                   // Insert the message length
     strcat(BufNum,"\r\n");                      // Need this for USART2
     USART2_PutQueue(BufNum);                    // Display AT message

     USART2_PutQueue(htmlPage);                  // Send web page data
     USART2_PutQueue("\r\n");                    // Need this for SerialUtility AP to show message
     USART2_PutQueue("AT+CIPCLOSE=0\r\n");       // Close the TCP connection
}

/**************************************************************************************************
  WiFiESP - Create UDP client and send data.
**************************************************************************************************/
void ClientUDP(void) {
     char *BufNum;

     if (CheckForValidIP(PC_IP_Address) == 0) {  // Make sure we have a valid IP address
         WIFI_DEBUGOUT("Diag: The IP is valid\r\n");
     } else {
         USB_PutQueue("Error: The IP is not valid\r\n");
         return;
     }

     WIFI_DEBUGOUT("Diag: Creating UDP client\r\n");
     // Connection to "Remote IP","Remote port","Local port", Remote IP and port are fixed
     //USART3_PutQueueText("AT+CIPSTART=\"UDP\",\"192.168.0.14\",1501,6000,0\r\n");
     BufNum = GetBuffer();
     sprintf(BufNum,"AT+CIPSTART=\"UDP\",\"");   // Start building AT command buffer
     strcat(BufNum,PC_IP_Address);               // Insert my PC IP address
     strcat(BufNum,"\",1501,6000,0\r\n");        // Insert remote port and local port
     USART3_PutQueueText(BufNum);                // Send the AT command
     LookForPrompt(portMAX_DELAY,"OK",1);        // Wait for prompt


     WIFI_DEBUGOUT("Diag: Setting data length\r\n");
     USART3_PutQueueText("AT+CIPSEND=10\r\n");   // Set data length in bytes including CRLF
     LookForPrompt(portMAX_DELAY,"OK",1);

     LookForPrompt(portMAX_DELAY,">",1);

     WIFI_DEBUGOUT("Diag: Sending data\r\n");
     WIFI_DEBUGOUT("Data: 001,test\r\n");
     USART3_PutQueueText("001,test\r\n");        // Send data
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Waiting for acknowledgement\r\n");
     LookForPrompt(portMAX_DELAY,"ACK,001",1);   // Check acknowledgement

     WIFI_DEBUGOUT("Diag: Closing the UDP connection\r\n");
     USART3_PutQueueText("AT+CIPCLOSE\r\n");     // Close the UDP connection
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  WiFiESP - Create UDP pass through mode.
**************************************************************************************************/
void PassthroughUDP(void) {
     char *BufNum;

     if (CheckForValidIP(PC_IP_Address) == 0) {  // Make sure we have a valid IP address
         WIFI_DEBUGOUT("Diag: The IP is valid\r\n");
     } else {
         USB_PutQueue("Error: The IP is not valid\r\n");
         return;
     }

     USB_PutQueue("Diag: Creating UDP client\r\n");
     // Connection to "Remote IPs","Remote port","Local port",Remote IP and port are fixed
     //USART3_PutQueueText("AT+CIPSTART=\"UDP\",\"192.168.0.14\",1501,6000,0\r\n");
     BufNum = GetBuffer();
     sprintf(BufNum,"AT+CIPSTART=\"UDP\",\"");   // Start building AT command buffer
     strcat(BufNum,PC_IP_Address);               // Insert my PC IP address
     strcat(BufNum,"\",1501,6000,0\r\n");        // Insert remote port and local port
     USART3_PutQueueText(BufNum);                // Send the AT command
     LookForPrompt(portMAX_DELAY,"OK",1);        // Wait for prompt

     WIFI_DEBUGOUT("Diag: Setting pass through mode\r\n");
     USART3_PutQueueText("AT+CIPMODE=1\r\n");    // Enable pass through mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: All data will be sent to remote\r\n");
     USART3_PutQueueText("AT+CIPSEND\r\n");
     LookForPrompt(portMAX_DELAY,"OK",1);

     LookForPrompt(portMAX_DELAY,">",1);

     WIFI_DEBUGOUT("Diag: Sending data\r\n");
     USART3_PutQueueText("\r\n");                // Starts a new line in the application
     USART3_PutQueueText("How now brown cow?\r\n");
     USART3_PutQueueText("Same way black bull.\r\n");
     vTaskDelay(pdMS_TO_TICKS(1000));

     WIFI_DEBUGOUT("Diag: Sending +++\r\n");     // Sending escape
     USART3_PutQueueText("+++");
     vTaskDelay(pdMS_TO_TICKS(1000));

     WIFI_DEBUGOUT("Diag: Returning to normal mode\r\n");
     USART3_PutQueueText("AT+CIPMODE=0\r\n");    // Disable pass through mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Closing the UDP connection\r\n");
     USART3_PutQueueText("AT+CIPCLOSE\r\n");     // Close the UDP connection
     LookForPrompt(portMAX_DELAY,"OK",1);
}

/**************************************************************************************************
  Construct DNS request.

    *************************************************
    *                  DNS HEADER                   *
    *************************************************

      0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                      ID                       |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |QR|   Opcode  |AA|TC|RD|RA|   Z    |   RCODE   |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    QDCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    ANCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    NSCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    ARCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+


   **************************************************
   *                  DNS QUESTION                  *
   **************************************************

      0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                                               |
    /                     QNAME                     /
    /                                               /
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                     QTYPE                     |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                     QCLASS                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

  This function creates a DNS header and question blocks using the URL pointed to by Name and
  saves it in QueryBuffer. The length of the query is saved in QueryBufLen.
  If the Name is NULL, the function exits with a return code of -1. If the string at Name is
  greater than 32 bytes, the function exits with a return code of -1. Otherwise, the function
  returns 0.
**************************************************************************************************/
int ConstructDNS(char *Name) {
    char *BufNum;
    int   NameSize;
    char  NameCopy[32]; // Increase if URL is longer than can fit in here
    char  Token[16];
    short Status;
    int   i, j;

    NameSize = strlen(Name);      // Get the length of the URL
    if (NameSize == 0) {
        USB_PutQueue("Error: Query name string is empty\r\n");
        return -1;
    } else {
#if WIFI_DEBUG
        BufNum = GetBuffer();     // Diagnostic output
        sprintf(BufNum,"Diag: URL name is %d characters long\r\n",NameSize);
        USB_PutQueue(BufNum);
#endif
    }

    memset(NameCopy,0,sizeof(NameCopy));    // Probably not necessary
    if (NameSize < sizeof(NameCopy)) {      // Check that the URL will fit in local buffer
        strcpy(NameCopy,Name);              // Copy of server name will be destroyed
#if WIFI_DEBUG
        BufNum = GetBuffer();     // Diagnostic output
        sprintf(BufNum,"Diag: URL name: %s\r\n",NameCopy);
        USB_PutQueue(BufNum);
#endif
    } else {                                // Error output
        USB_PutQueue("Error: Query name won't fit into available buffer\r\n");
        return -1;
    }

    memset(QueryBuffer,0,sizeof(QueryBuffer));
    i = 0;                                  // i will count the number of bytes in the message
    QueryBuffer[i++] = 0x56;                // 'V'
    QueryBuffer[i++] = 0x47;                // 'G'
    QueryBuffer[i++] = 0x01;                // FLAGS
    QueryBuffer[i++] = 0x00;                //   "
    QueryBuffer[i++] = 0x00;                // QDCOUNT
    QueryBuffer[i++] = 0x01;                //   "
    QueryBuffer[i++] = 0x00;                // ANCOUNT
    QueryBuffer[i++] = 0x00;                //   "
    QueryBuffer[i++] = 0x00;                // NSCOUNT
    QueryBuffer[i++] = 0x00;                //   "
    QueryBuffer[i++] = 0x00;                // AECOUNT
    QueryBuffer[i++] = 0x00;                //   "

    do {                                    // Get all parts of the URL and build query buffer
        memset(Token,0,sizeof(Token));
        Status = GetToken(Token,NameCopy,'.');
        QueryBuffer[i++] = strlen(Token);
        for (j=0;j<strlen(Token);j++) {
            QueryBuffer[i++] = Token[j];
        }
    } while (Status > 0);                   // No more tokens when status is 0
    QueryBuffer[i++] = 0x00;                // NULL

    QueryBuffer[i++] = 0x00;                // QTYPE
    QueryBuffer[i++] = 0x01;                //   "
    QueryBuffer[i++] = 0x00;                // QCLASS
    QueryBuffer[i++] = 0x01;                //   "
#if WIFI_DEBUG
    HexPrint(&QueryBuffer[0],12);           // Display the header
    HexPrint(&QueryBuffer[12],i-12);        // Display the question

    BufNum = GetBuffer();
    sprintf(BufNum,"Diag: Length of query buffer: %d\r\n",i);
    USB_PutQueue(BufNum);
#endif
    QueryBufLen = i;                        // Save the message length
#if WIFI_DEBUG
    DispDNS_QueryBuf();                     // Diagnostic display
#endif
    return 0;
}

/**************************************************************************************************
  WiFiESP - Make a request to DNS server. DNS servers listen on port 53.
  USART3_PutQueueText("AT+CIPSTART=\"UDP\",\"208.67.222.222\",53,6000,0\r\n"); //
  USART3_PutQueueText("AT+CIPSTART=\"UDP\",\"8.8.8.8\",53,6000,0\r\n");        // Google IP
  USART3_PutQueueText("AT+CIPSTART=\"UDP\",\"64.71.255.204\",53,6000,0\r\n");  // Rogers IP
  USART3_PutQueueText("AT+CIPSTART=\"UDP\",\"192.168.0.15\",1501,6000,0\r\n"); // Local IP
**************************************************************************************************/
void RequestDNS(void) {
     char *BufNum;

     if (CheckForValidIP(DNS_SErverIP) == 0) {   // Make sure we have a valid IP address
         WIFI_DEBUGOUT("Diag: The IP is valid\r\n");
     } else {
         USB_PutQueue("Error: The IP is not valid\r\n");
         return;
     }

     ConstructDNS(NTP_ServerName);

     // Connection to "IP address","Remote port","Local port",Remote IP and port are fixed
     WIFI_DEBUGOUT("Diag: Creating UDP client\r\n");
     sprintf(UDP_Buffer,"AT+CIPSTART=\"UDP\",\"");
     strcat(UDP_Buffer,DNS_SErverIP);
     strcat(UDP_Buffer,"\",53,6000,0\r\n");
     WIFI_DEBUGOUT(UDP_Buffer);

#if WIFI_DEBUG
     BufNum = GetBuffer();        // Diagnostic
     sprintf(BufNum,"Diag: UDP message length: %d\r\n",strlen(UDP_Buffer));
     USB_PutQueue(BufNum);
#endif
     USART3_PutQueueText(UDP_Buffer);
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Setting pass through mode\r\n");
     USART3_PutQueueText("AT+CIPMODE=1\r\n");   // Enable pass through mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: All data will be sent to remote\r\n");
     USART3_PutQueueText("AT+CIPSEND\r\n");     // Start sending data in transparent mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     LookForPrompt(portMAX_DELAY,">",1);

     Transparent = 1;        // Set USART3 to transparent mode
     U3RcvBufPtr = 0;        // Counter will increment for each character received
     memset(U3RcvBuf,0,U3_RCV_BUF_SIZE);

     /********************************************************************************************/
     /* Diagnostic
     BufNum = GetBuffer();
     sprintf(BufNum,"Diag: Size of DNS query: %d characters\r\n",QueryBufLen);
     USB_PutQueue(BufNum);
     */

     WIFI_DEBUGOUT("Diag: Sending DNS message\r\n");
     USART3_PutQueueBinary(QueryBuffer,QueryBufLen);  // Send DNS query message

     /* Diagnostic
     BufNum = GetBuffer();
     sprintf(BufNum,"Diag: USART3 receive buffer pointer: %d characters\r\n",U3RcvBufPtr);
     USB_PutQueue(BufNum);
     */
     /********************************************************************************************/

     WIFI_DEBUGOUT("Diag: Waiting for response\r\n");
                                            // Don't know the length of response
     vTaskDelay(pdMS_TO_TICKS(1000));       // Have to use timer

#if WIFI_DEBUG
     BufNum = GetBuffer();                  // Diagnostic
     sprintf(BufNum,"Diag: Received %d characters\r\n",U3RcvBufPtr);
     USB_PutQueue(BufNum);
#endif

     memset(DNS_Answer,0,256);              // Clear answer buffer
     memcpy(DNS_Answer,U3RcvBuf,256);       // Copy from UART buffer to answer buffer

     vTaskDelay(pdMS_TO_TICKS(1000));       // Delay 1 second
     WIFI_DEBUGOUT("Diag: Sending +++\r\n");
     USART3_PutQueueBinary("+++",3);        // Take WiFiESP out of pass through mode
     vTaskDelay(pdMS_TO_TICKS(1000));       // Delay 1 second

     U3RcvBufPtr = 0;
     Transparent = 0;                       // Set USART3 to normal ASCII mode
     WIFI_DEBUGOUT("Diag: Returning to normal mode\r\n");
     USART3_PutQueueText("AT+CIPMODE=0\r\n");   // Disable pass through mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Closing the UDP connection\r\n");
     USART3_PutQueueText("AT+CIPCLOSE\r\n");    // Close the UDP connection
     LookForPrompt(portMAX_DELAY,"OK",1);

     DispDNSanswerBuf();                    // Print answer buffer
     ParseDNS(&DNS_Answer);                 // Parse answer buffer
}

/**************************************************************************************************
  Parse DNS answer.

    *************************************************
    *                  DNS HEADER                   *
    *************************************************

      0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                      ID                       |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |QR|   Opcode  |AA|TC|RD|RA|   Z    |   RCODE   |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    QDCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    ANCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    NSCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    ARCOUNT                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+


   **************************************************
   *                  DNS ANSWER                    *
   **************************************************

      0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                                               |
    /                     NAME                      /
    /                                               /
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                     TYPE                      |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                     CLASS                     |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                      TTL                      |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                    RDLENGTH                   |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
    |                      RDATA                    |
    +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

**************************************************************************************************/
int ParseDNS(char *answer) {
    char *BufNum;
    char  RtnCode;
    int   NumIPs;
    char  SegName[32];
    int   SegSize;
    char  Addr[16];
    int   AddrSize;
    int   i, j, k;
    //
    if ((answer[0] != QueryBuffer[0]) || (answer[1] != QueryBuffer[1])) {
        USB_PutQueue("Error: Answer name not equal to query name\r\n");
        return -1;
    }
    //
    RtnCode = answer[3] & 0x0F;  // Have to check the return code
    switch (RtnCode) {
        case '\0':  // No error
            break;
        case '\1':
            USB_PutQueue("Error: Format error\r\n");
            break;
        case '\2':
            USB_PutQueue("Error: Server failure\r\n");
            break;
        case '\3':
            USB_PutQueue("Error: Name error\r\n");
            break;
        case '\5':
            USB_PutQueue("Error: Refused\r\n");
            break;
    }

    if (RtnCode == 0) {  // No sense doing any of this if there was an error
        NumIPs = answer[7];
#if WIFI_DEBUG
        BufNum = GetBuffer(); // Diagnostic
        sprintf(BufNum,"Diag: Number of IPs returned: %d\r\n",NumIPs);
        USB_PutQueue(BufNum);
#endif
        i = 12; // Advance index to start of requested server name
        do {
            memset(SegName,0,sizeof(SegName));
            SegSize = answer[i++];              // Get size of name token

            if (SegSize > 0) {                  // NULL indicates end of name
#if WIFI_DEBUG
                BufNum = GetBuffer();  // Diagnostic
                sprintf(BufNum,"Diag: Segment has %d characters\r\n",SegSize);
                USB_PutQueue(BufNum);
#endif
                for (j=0;j<SegSize;j++) {
                    SegName[j] = answer[i++];   // Move name token to local buffer
                }
#if WIFI_DEBUG
                BufNum = GetBuffer();  // Diagnostic
                strcpy(BufNum,SegName);
                strcat(BufNum,"\r\n");
                USB_PutQueue(BufNum);
#endif
            }
        } while (SegSize > 0);         // Exit when all name tokens processed
#if WIFI_DEBUG
        BufNum = GetBuffer();          // Diagnostic
        sprintf(BufNum,"Diag: Value of i: %d\r\n",i);
        USB_PutQueue(BufNum);
#endif
        i += 15;  // Advance index to size value of first IP address field
        k = 0;    // Index counter of IP_Address (wraps after 4 addresses)
        while (NumIPs > 0) {
            AddrSize = answer[i++];    // Get address size (usually 0x04)
#if WIFI_DEBUG
            BufNum = GetBuffer();      // Diagnostic
            sprintf(BufNum,"Diag: IP field has %d characters\r\n",AddrSize);
            USB_PutQueue(BufNum);
#endif
            memset(Addr,0,sizeof(Addr));
            for (j=0;j<AddrSize;j++) {
                Addr[j] = answer[i++]; // Get IP address characters
            }

            // Combine hex characters into decimal with dots between elements
            memset(IP_Address[k],0,sizeof(IP_Address[k]));
            for (j=0;j<4;j++) {
                sprintf(IP_Address[k],"%d.%d.%d.%d",Addr[0],Addr[1],Addr[2],Addr[3]);
            }
#if WIFI_DEBUG
            BufNum = GetBuffer();      // Diagnostic
            strcpy(BufNum,IP_Address[k]);
            strcat(BufNum,"\r\n");
            USB_PutQueue(BufNum);
#endif
            i += 11;          // Advance index to size value of next IP address field
            NumIPs -= 1;
            k += 1;
            if (k > 3) k = 0; // Wrap IP address buffer location if received more that 4
        }
        return 0;
    } else {
        return -1;
    }
}

/**************************************************************************************************
  WiFiESP - Construct NTP packet.
**************************************************************************************************/
void ConstructNTP(void) {
     char *BufNum;

     memset(NTP_Query,0,sizeof(NTP_Query));
     NTP_Query[0]  = 0b11100011;
     NTP_Query[1]  = 0;
     NTP_Query[2]  = 6;
     NTP_Query[3]  = 0xEC;

     NTP_Query[12] = 49;
     NTP_Query[13] = 0x4E;
     NTP_Query[14] = 49;
     NTP_Query[15] = 52;

     //DispNTP_QueryBuf();
}

/**************************************************************************************************
  WiFiESP - Make a request to NTP server. NTP servers listen on port 123.
  ntp-b.nist.gov     132.163.96.5
  ntp-www.nist.gov   132.163.97.5
  ntp-c.colorado.edu 128.138.141.177
  ntp-d.nist.gov     129.6.15.32
**************************************************************************************************/
void RequestNTP(void) {
     char *BufNum;

     ConstructNTP();

     // Connection to "IP address","Remote port","Local port",Remote IP and port are fixed
     WIFI_DEBUGOUT("Diag: Creating UDP client\r\n");
     //USART3_PutQueueText("AT+CIPSTART=\"UDP\",\"132.246.11.238\",123,6000,0\r\n");
     BufNum = GetBuffer();
     sprintf(BufNum,"AT+CIPSTART=\"UDP\",\"");
     strcat(BufNum,NTP_ServerIP);
     strcat(BufNum,"\",123,6000,0\r\n");
     USART3_PutQueueText(BufNum);
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Setting pass through mode\r\n");
     USART3_PutQueueText("AT+CIPMODE=1\r\n");   // Enable pass through mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: All data will be sent to remote\r\n");
     USART3_PutQueueText("AT+CIPSEND\r\n");
     LookForPrompt(portMAX_DELAY,">",1);

     Transparent = 1;        // Set USART3 to transparent mode
     U3RcvBufPtr = 0;        // Counter will increment for each character received
     memset(U3RcvBuf,0,256);

     //XmtByteCntr = sizeof(NTP_Query);  // Have to know the number of characters to transmit
     WIFI_DEBUGOUT("Diag: Sending NTP message\r\n");
     USART3_PutQueueBinary(NTP_Query,sizeof(NTP_Query));

     WIFI_DEBUGOUT("Diag: Waiting for response\r\n");
     //while (U3RcvBufPtr<48);
     //while (U3RcvBufPtr<sizeof(NTP_Query); // NTP_Query and Answer should be 48 characters
     vTaskDelay(pdMS_TO_TICKS(1000));  // Use timer instead

     memset(NTP_Answer,0,SIZE_OF_NTP_BUF);
     memcpy(NTP_Answer,U3RcvBuf,SIZE_OF_NTP_BUF);

     vTaskDelay(pdMS_TO_TICKS(1000));
     WIFI_DEBUGOUT("Diag: Sending +++\r\n");
     USART3_PutQueueBinary("+++",3);
     vTaskDelay(pdMS_TO_TICKS(1000));

#if WIFI_DEBUG
     BufNum = GetBuffer();
     sprintf(BufNum,"Diag: Received %d characters\r\n",U3RcvBufPtr);
     USB_PutQueue(BufNum);
#endif

     U3RcvBufPtr = 0;
     Transparent = 0;        // Set USART3 to normal ASCII mode
     WIFI_DEBUGOUT("Diag: Returning to normal mode\r\n");
     USART3_PutQueueText("AT+CIPMODE=0\r\n");   // Disable pass through mode
     LookForPrompt(portMAX_DELAY,"OK",1);

     WIFI_DEBUGOUT("Diag: Closing the UDP connection\r\n");
     USART3_PutQueueText("AT+CIPCLOSE\r\n");    // Close the UDP connection
     LookForPrompt(portMAX_DELAY,"OK",1);

     DispNTPAnswerBuf();
     ParseNTP(&NTP_Answer);
}

/**************************************************************************************************
  WiFiESP - Parse NTP answer.
**************************************************************************************************/
void ParseNTP(char *Name) {
     char *BufNum;
     unsigned long NTP_Time;

     NTP_Time = (NTP_Answer[40]<<24) | (NTP_Answer[41]<<16) | (NTP_Answer[42]<<8) | NTP_Answer[43];
     BufNum = GetBuffer();
     sprintf(BufNum,"Seconds since Jan 1, 1900: %04X\r\n",NTP_Time);
     USB_PutQueue(BufNum);
     Highest(epoch) = NTP_Answer[40];
     Higher(epoch)  = NTP_Answer[41];
     Hi(epoch)      = NTP_Answer[42];
     Lo(epoch)      = NTP_Answer[43];

     epoch = epoch - 2208988800 - 14400;
     (void) Time_epochToDate(epoch, &my_time);

     BufNum = GetBuffer();
     sprintf(BufNum,"Time: %d/%02d/%02d %02d:%02d:%02d\r\n",my_time.yy,my_time.mo,my_time.md,my_time.hh,my_time.mn,my_time.ss);
     USB_PutQueue(BufNum);

     MyTime.seconds = my_time.ss;
     MyTime.minutes = my_time.mn;
     MyTime.hours   = my_time.hh;
     MyTime.day     = my_time.md;
     MyTime.month   = my_time.mo;
     MyTime.year    = my_time.yy;
     RTC_Write(&MyTime);
     GetTime();
}

/**************************************************************************************************
  WiFiESP - Check command queue and execute the command is one is found.
**************************************************************************************************/
void CheckCmdQueue(long Timeout) {
     BaseType_t rtnCode;      // Function return code
     char Command;
     char *BufNum;

     rtnCode = xQueueReceive(CmdQueue, &Command, Timeout); // Block here for Timeout milliseconds
     if (rtnCode == 1) {
#if WIFI_DEBUG
         BufNum = GetBuffer();
         sprintf(BufNum,"Diag: Received command %c from queue\r\n",Command);
         USB_PutQueue(BufNum);
#endif
         switch (Command) {
             case 'A':
                 // Not used
                 break;
             case 'B':
                 WiFiESP_SW_Reset();
                 break;
             case 'C':
                 WiFiESP_Set_Factory_Defaults();
                 break;
             case 'D':
                 WiFiESP_Configure_Flash();
                 break;
             case 'E':
                 WiFiESP_Set_UART_Config();
                 break;
             case 'F':
                 SendAT();
                 break;
             case 'G':
                 RequestDNS();
                 break;
             case 'H':
                 RequestNTP();
                 break;
             case 'I':
                 ServerTCPforHTTP();
                 break;
             case 'J':
                 ClientUDP();
                 break;
             case 'K':
                 ClientTCP();
                 break;
             case 'L':
                 PassthroughUDP();
                 break;
             case 'M':
                 ListAvailAPs();
                 break;
         }
     }
}

/**************************************************************************************************
  WiFiESP - Put a command in the queue.
            This function enters a command to be executed by the WiFiESP_Task into the command
            queue. The valid commands are defined in WiFi_ESP_Driver.h. These can be expanded as
            the need arises.
**************************************************************************************************/
void WiFiESP_PutQueue(char cmd) {
     BaseType_t rtnCode;
     char Command;
     char *BufNum;

     Command = cmd;
     rtnCode = xQueueSend(CmdQueue, &Command, 0); // No blocking

     if (rtnCode == pdFAIL) {
         USB_PutQueue("Error: Failed to enter command into WiFiESP queue\r\n");
     } else {
#if WIFI_DEBUG
         BufNum = GetBuffer();
         sprintf(BufNum,"Diag: Entered command %c into queue\r\n",Command);
         USB_PutQueue(BufNum);
#endif
     }


}

/**************************************************************************************************
  WiFiESP Task. This is the WiFiESP task. It sets up the queues and then checks for unsolicited
                messages from the ESP chip for a few milliseconds and then checks the command
                queue for commands to execute.
**************************************************************************************************/
void WiFiESP_Task(void *pvParameters) {

     // Create a queue of pointers
     vWiFiESPPointerQueue = xQueueCreate(64, sizeof(char *));
     if (vWiFiESPPointerQueue == NULL) {
         USB_PutQueue("Error: Could not create WiFi pointer queue\r\n");
     } else {
         USB_PutQueue("Diag: WiFi pointer queue created successfully\r\n");
         vQueueAddToRegistry(vWiFiESPPointerQueue, "WiFiESP message pointer queue");
     }

     // Create a queue of character
     CmdQueue = xQueueCreate(16, sizeof(char));
     if (CmdQueue == NULL) {
         USB_PutQueue("Error: Could not create command queue\r\n");
     } else {
         USB_PutQueue("Diag: WiFi command queue created successfully\r\n");
         vQueueAddToRegistry(CmdQueue, "WiFiESP command queue");
     }

     USB_PutQueue("Diag: WiFiESP task started\r\n");

     // This is not used for now
     /* Create a binary semaphore
     vWiFiESPSemHandle = xSemaphoreCreateBinary();
     if (vWiFiESPSemHandle == NULL) {
         USB_PutQueue("Could not create WiFi semaphore\r\n");
     } else {
         USB_PutQueue("WiFi semaphore created successfully\r\n");
         vQueueAddToRegistry(vWiFiESPSemHandle, "WiFiESP Semaphore");
     }
     */

     //
     WiFiESP_Initialize();
     vTaskDelay(pdMS_TO_TICKS(18000)); // Big delay to allow USB to enumerate
     WiFiESP_HW_Reset();
     WiFiESP_Status = 1;
     vTaskDelay(pdMS_TO_TICKS(500));
     //
     //WiFiESP_SW_Reset();               // Not required

     WaitForIP();                      // The ESP chip will get an IP address from DHCP

     //WiFiESP_PutQueue(REQUEST_DNS_SERVICE);
     //WiFiESP_PutQueue(REQUEST_NTP_SERVICE);



     while (1) {

         WaitForResponse(100); // Wait for and display any unsolicited Messages
         CheckCmdQueue(100);   // Execute any WiFi commands that come in
         //vTaskDelay(pdMS_TO_TICKS(1000));
     }
}








/**************************************************************************************************
  Display DNS query buffer.
**************************************************************************************************/
void DispDNS_QueryBuf(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",QueryBuffer[0],QueryBuffer[1],
     QueryBuffer[2],QueryBuffer[3],QueryBuffer[4],QueryBuffer[5],QueryBuffer[6],QueryBuffer[7]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",QueryBuffer[8],QueryBuffer[9],
     QueryBuffer[10],QueryBuffer[11],QueryBuffer[12],QueryBuffer[13],QueryBuffer[14],QueryBuffer[15]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",QueryBuffer[16],QueryBuffer[17],
     QueryBuffer[18],QueryBuffer[19],QueryBuffer[20],QueryBuffer[21],QueryBuffer[22],QueryBuffer[23]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",QueryBuffer[24],QueryBuffer[25],
     QueryBuffer[26],QueryBuffer[27],QueryBuffer[28],QueryBuffer[29],QueryBuffer[30],QueryBuffer[31]);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  Display DNS answer buffer.
**************************************************************************************************/
void DispDNSanswerBuf(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[0],DNS_Answer[1],
     DNS_Answer[2],DNS_Answer[3],DNS_Answer[4],DNS_Answer[5],DNS_Answer[6],DNS_Answer[7]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[8],DNS_Answer[9],
     DNS_Answer[10],DNS_Answer[11],DNS_Answer[12],DNS_Answer[13],DNS_Answer[14],DNS_Answer[15]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[16],DNS_Answer[17],
     DNS_Answer[18],DNS_Answer[19],DNS_Answer[20],DNS_Answer[21],DNS_Answer[22],DNS_Answer[23]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[24],DNS_Answer[25],
     DNS_Answer[26],DNS_Answer[27],DNS_Answer[28],DNS_Answer[29],DNS_Answer[30],DNS_Answer[31]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[32],DNS_Answer[33],
     DNS_Answer[34],DNS_Answer[35],DNS_Answer[36],DNS_Answer[37],DNS_Answer[38],DNS_Answer[39]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[40],DNS_Answer[41],
     DNS_Answer[42],DNS_Answer[43],DNS_Answer[44],DNS_Answer[45],DNS_Answer[46],DNS_Answer[47]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[48],DNS_Answer[49],
     DNS_Answer[50],DNS_Answer[51],DNS_Answer[52],DNS_Answer[53],DNS_Answer[54],DNS_Answer[55]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[56],DNS_Answer[57],
     DNS_Answer[58],DNS_Answer[59],DNS_Answer[60],DNS_Answer[61],DNS_Answer[62],DNS_Answer[63]);
     USB_PutQueue(BufNum);

/*************************************************************************************************/

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[64],DNS_Answer[65],
     DNS_Answer[66],DNS_Answer[67],DNS_Answer[68],DNS_Answer[69],DNS_Answer[70],DNS_Answer[71]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[72],DNS_Answer[73],
     DNS_Answer[74],DNS_Answer[75],DNS_Answer[76],DNS_Answer[77],DNS_Answer[78],DNS_Answer[79]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[80],DNS_Answer[81],
     DNS_Answer[82],DNS_Answer[83],DNS_Answer[84],DNS_Answer[85],DNS_Answer[86],DNS_Answer[87]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[88],DNS_Answer[89],
     DNS_Answer[90],DNS_Answer[91],DNS_Answer[92],DNS_Answer[93],DNS_Answer[94],DNS_Answer[95]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[96],DNS_Answer[97],
     DNS_Answer[98],DNS_Answer[99],DNS_Answer[100],DNS_Answer[101],DNS_Answer[102],DNS_Answer[103]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[104],DNS_Answer[105],
     DNS_Answer[106],DNS_Answer[107],DNS_Answer[108],DNS_Answer[109],DNS_Answer[110],DNS_Answer[111]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[112],DNS_Answer[113],
     DNS_Answer[114],DNS_Answer[115],DNS_Answer[116],DNS_Answer[117],DNS_Answer[118],DNS_Answer[119]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[120],DNS_Answer[121],
     DNS_Answer[122],DNS_Answer[123],DNS_Answer[124],DNS_Answer[125],DNS_Answer[126],DNS_Answer[127]);
     USB_PutQueue(BufNum);

/*************************************************************************************************/

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[128],DNS_Answer[129],
     DNS_Answer[130],DNS_Answer[131],DNS_Answer[132],DNS_Answer[133],DNS_Answer[134],DNS_Answer[135]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[136],DNS_Answer[137],
     DNS_Answer[138],DNS_Answer[139],DNS_Answer[140],DNS_Answer[141],DNS_Answer[142],DNS_Answer[143]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[144],DNS_Answer[145],
     DNS_Answer[146],DNS_Answer[147],DNS_Answer[148],DNS_Answer[149],DNS_Answer[150],DNS_Answer[151]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[152],DNS_Answer[153],
     DNS_Answer[154],DNS_Answer[155],DNS_Answer[156],DNS_Answer[157],DNS_Answer[158],DNS_Answer[159]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[160],DNS_Answer[161],
     DNS_Answer[162],DNS_Answer[163],DNS_Answer[164],DNS_Answer[165],DNS_Answer[166],DNS_Answer[167]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[168],DNS_Answer[169],
     DNS_Answer[170],DNS_Answer[171],DNS_Answer[172],DNS_Answer[173],DNS_Answer[174],DNS_Answer[175]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[176],DNS_Answer[177],
     DNS_Answer[178],DNS_Answer[179],DNS_Answer[180],DNS_Answer[181],DNS_Answer[182],DNS_Answer[183]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[184],DNS_Answer[185],
     DNS_Answer[186],DNS_Answer[187],DNS_Answer[188],DNS_Answer[189],DNS_Answer[190],DNS_Answer[191]);
     USB_PutQueue(BufNum);

/*************************************************************************************************/

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[192],DNS_Answer[193],
     DNS_Answer[194],DNS_Answer[195],DNS_Answer[196],DNS_Answer[197],DNS_Answer[198],DNS_Answer[199]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[200],DNS_Answer[201],
     DNS_Answer[202],DNS_Answer[203],DNS_Answer[204],DNS_Answer[205],DNS_Answer[206],DNS_Answer[207]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[208],DNS_Answer[209],
     DNS_Answer[210],DNS_Answer[211],DNS_Answer[212],DNS_Answer[213],DNS_Answer[214],DNS_Answer[215]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[216],DNS_Answer[217],
     DNS_Answer[218],DNS_Answer[219],DNS_Answer[220],DNS_Answer[221],DNS_Answer[222],DNS_Answer[223]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[224],DNS_Answer[225],
     DNS_Answer[226],DNS_Answer[227],DNS_Answer[228],DNS_Answer[229],DNS_Answer[230],DNS_Answer[231]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[232],DNS_Answer[233],
     DNS_Answer[234],DNS_Answer[235],DNS_Answer[236],DNS_Answer[237],DNS_Answer[238],DNS_Answer[239]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[240],DNS_Answer[241],
     DNS_Answer[242],DNS_Answer[243],DNS_Answer[244],DNS_Answer[245],DNS_Answer[246],DNS_Answer[247]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",DNS_Answer[248],DNS_Answer[249],
     DNS_Answer[250],DNS_Answer[251],DNS_Answer[252],DNS_Answer[253],DNS_Answer[254],DNS_Answer[255]);
     USB_PutQueue(BufNum);
}

/**************************************************************************************************
  Display NTP query buffer.
**************************************************************************************************/
void DispNTP_QueryBuf(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Query[0],NTP_Query[1],
     NTP_Query[2],NTP_Query[3],NTP_Query[4],NTP_Query[5],NTP_Query[6],NTP_Query[7]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Query[8],NTP_Query[9],
     NTP_Query[10],NTP_Query[11],NTP_Query[12],NTP_Query[13],NTP_Query[14],NTP_Query[15]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Query[16],NTP_Query[17],
     NTP_Query[18],NTP_Query[19],NTP_Query[20],NTP_Query[21],NTP_Query[22],NTP_Query[23]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Query[24],NTP_Query[25],
     NTP_Query[26],NTP_Query[27],NTP_Query[28],NTP_Query[29],NTP_Query[30],NTP_Query[31]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Query[32],NTP_Query[33],
     NTP_Query[34],NTP_Query[35],NTP_Query[36],NTP_Query[37],NTP_Query[38],NTP_Query[39]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Query[40],NTP_Query[41],
     NTP_Query[42],NTP_Query[43],NTP_Query[44],NTP_Query[45],NTP_Query[46],NTP_Query[47]);
     USB_PutQueue(BufNum);
}


/**************************************************************************************************
  Display NTP answer buffer.
**************************************************************************************************/
void DispNTPAnswerBuf(void) {
     char *BufNum;

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[0],NTP_Answer[1],
     NTP_Answer[2],NTP_Answer[3],NTP_Answer[4],NTP_Answer[5],NTP_Answer[6],NTP_Answer[7]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[8],NTP_Answer[9],
     NTP_Answer[10],NTP_Answer[11],NTP_Answer[12],NTP_Answer[13],NTP_Answer[14],NTP_Answer[15]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[16],NTP_Answer[17],
     NTP_Answer[18],NTP_Answer[19],NTP_Answer[20],NTP_Answer[21],NTP_Answer[22],NTP_Answer[23]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[24],NTP_Answer[25],
     NTP_Answer[26],NTP_Answer[27],NTP_Answer[28],NTP_Answer[29],NTP_Answer[30],NTP_Answer[31]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[32],NTP_Answer[33],
     NTP_Answer[34],NTP_Answer[35],NTP_Answer[36],NTP_Answer[37],NTP_Answer[38],NTP_Answer[39]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[40],NTP_Answer[41],
     NTP_Answer[42],NTP_Answer[43],NTP_Answer[44],NTP_Answer[45],NTP_Answer[46],NTP_Answer[47]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[48],NTP_Answer[49],
     NTP_Answer[50],NTP_Answer[51],NTP_Answer[52],NTP_Answer[53],NTP_Answer[54],NTP_Answer[55]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[56],NTP_Answer[57],
     NTP_Answer[58],NTP_Answer[59],NTP_Answer[60],NTP_Answer[61],NTP_Answer[62],NTP_Answer[63]);
     USB_PutQueue(BufNum);

/*************************************************************************************************/
     /*
     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[64],NTP_Answer[65],
     NTP_Answer[66],NTP_Answer[67],NTP_Answer[68],NTP_Answer[69],NTP_Answer[70],NTP_Answer[71]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[72],NTP_Answer[73],
     NTP_Answer[74],NTP_Answer[75],NTP_Answer[76],NTP_Answer[77],NTP_Answer[78],NTP_Answer[79]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[80],NTP_Answer[81],
     NTP_Answer[82],NTP_Answer[83],NTP_Answer[84],NTP_Answer[85],NTP_Answer[86],NTP_Answer[87]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[88],NTP_Answer[89],
     NTP_Answer[90],NTP_Answer[91],NTP_Answer[92],NTP_Answer[93],NTP_Answer[94],NTP_Answer[95]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[96],NTP_Answer[97],
     NTP_Answer[98],NTP_Answer[99],NTP_Answer[100],NTP_Answer[101],NTP_Answer[102],NTP_Answer[103]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[104],NTP_Answer[105],
     NTP_Answer[106],NTP_Answer[107],NTP_Answer[108],NTP_Answer[109],NTP_Answer[110],NTP_Answer[111]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[112],NTP_Answer[113],
     NTP_Answer[114],NTP_Answer[115],NTP_Answer[116],NTP_Answer[117],NTP_Answer[118],NTP_Answer[119]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[120],NTP_Answer[121],
     NTP_Answer[122],NTP_Answer[123],NTP_Answer[124],NTP_Answer[125],NTP_Answer[126],NTP_Answer[127]);
     USB_PutQueue(BufNum);
     */
/*************************************************************************************************/
     /*
     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[128],NTP_Answer[129],
     NTP_Answer[130],NTP_Answer[131],NTP_Answer[132],NTP_Answer[133],NTP_Answer[134],NTP_Answer[135]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[136],NTP_Answer[137],
     NTP_Answer[138],NTP_Answer[139],NTP_Answer[140],NTP_Answer[141],NTP_Answer[142],NTP_Answer[143]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[144],NTP_Answer[145],
     NTP_Answer[146],NTP_Answer[147],NTP_Answer[148],NTP_Answer[149],NTP_Answer[150],NTP_Answer[151]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[152],NTP_Answer[153],
     NTP_Answer[154],NTP_Answer[155],NTP_Answer[156],NTP_Answer[157],NTP_Answer[158],NTP_Answer[159]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[160],NTP_Answer[161],
     NTP_Answer[162],NTP_Answer[163],NTP_Answer[164],NTP_Answer[165],NTP_Answer[166],NTP_Answer[167]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[168],NTP_Answer[169],
     NTP_Answer[170],NTP_Answer[171],NTP_Answer[172],NTP_Answer[173],NTP_Answer[174],NTP_Answer[175]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[176],NTP_Answer[177],
     NTP_Answer[178],NTP_Answer[179],NTP_Answer[180],NTP_Answer[181],NTP_Answer[182],NTP_Answer[183]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[184],NTP_Answer[185],
     NTP_Answer[186],NTP_Answer[187],NTP_Answer[188],NTP_Answer[189],NTP_Answer[190],NTP_Answer[191]);
     USB_PutQueue(BufNum);
     */
/*************************************************************************************************/
     /*
     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[192],NTP_Answer[193],
     NTP_Answer[194],NTP_Answer[195],NTP_Answer[196],NTP_Answer[197],NTP_Answer[198],NTP_Answer[199]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[200],NTP_Answer[201],
     NTP_Answer[202],NTP_Answer[203],NTP_Answer[204],NTP_Answer[205],NTP_Answer[206],NTP_Answer[207]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[208],NTP_Answer[209],
     NTP_Answer[210],NTP_Answer[211],NTP_Answer[212],NTP_Answer[213],NTP_Answer[214],NTP_Answer[215]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[216],NTP_Answer[217],
     NTP_Answer[218],NTP_Answer[219],NTP_Answer[220],NTP_Answer[221],NTP_Answer[222],NTP_Answer[223]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[224],NTP_Answer[225],
     NTP_Answer[226],NTP_Answer[227],NTP_Answer[228],NTP_Answer[229],NTP_Answer[230],NTP_Answer[231]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[232],NTP_Answer[233],
     NTP_Answer[234],NTP_Answer[235],NTP_Answer[236],NTP_Answer[237],NTP_Answer[238],NTP_Answer[239]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[240],NTP_Answer[241],
     NTP_Answer[242],NTP_Answer[243],NTP_Answer[244],NTP_Answer[245],NTP_Answer[246],NTP_Answer[247]);
     USB_PutQueue(BufNum);

     BufNum = GetBuffer();
     sprintf(BufNum,"%02X %02X %02X %02X %02X %02X %02X %02X\r\n",NTP_Answer[248],NTP_Answer[249],
     NTP_Answer[250],NTP_Answer[251],NTP_Answer[252],NTP_Answer[253],NTP_Answer[254],NTP_Answer[255]);
     USB_PutQueue(BufNum);
     */
}












/**************************************************************************************************
  End of File
**************************************************************************************************/