/**************************************************************************************************
  WiFi_ESP_Driver.h
  Defines a pool of buffers for printing messages.


  Copyright (C) 2007  Val Gretchev

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************************************/
#ifndef WIFIESP_H
#define WIFIESP_H


// Defines
#define NUM_MESSAGES 20
#define MESSAGE_SIZE 256
#define MAX_WIFI_MSG_SIZE 2048
#define IP_ADDR_SIZE 16
#define SIZE_OF_NTP_BUF 64

#define SEND_SW_RST 'B'
#define SET_FACT_DEFAULTS 'C'
#define CONFIG_FLASH 'D'
#define SET_UART_CONFIG 'E'
#define SEND_AT 'F'
#define REQUEST_DNS_SERVICE 'G'
#define REQUEST_NTP_SERVICE 'H'
#define START_HTTP_SERVER 'I'
#define START_UDP_CLIENT 'J'
#define START_TCP_CLIENT 'K'
#define START_UDP_PASSTHROUGH 'L'
#define LIST_AVAIL_APS 'M'

// Function Prototypes
short CheckForValidIP(char *Addr);
short GetToken(char *t, char *s, char c);
void ByteCopy(unsigned char *block, unsigned char *dest, int size);
int HexPrint( unsigned char *array, int size);

void WiFiESP_Initialize(void);
void WiFiESP_HW_Reset(void);
void WiFiESP_SW_Reset(void);
void WiFiESP_Set_Factory_Defaults(void);
void WiFiESP_Configure_Flash(void);
void WiFiESP_Set_UART_Config(void);
void SendAT(void);
void ListAvailAPs(void);
char WaitForResponse(unsigned long Timeout);
short LookForPrompt(unsigned long Timeout, char *Prompt, char RstMsgCntr);
void WaitForReady(void);
void WaitForIP(void);


void GetLocalIP(void);
void ClientTCP(void);
short SendLength(char *Msg);
short SendLongMessage(const code char *Msg);
void ServerTCPforHTTP(void);
void TestForHTTP(void);
void ClientUDP(void);
void PassthroughUDP(void);
int  ConstructDNS(char *Name);
void MakeUDPforDNS(void);
void RequestDNS(void);
int  ParseDNS(char *answer);
void ConstructNTP(void);
void RequestNTP(void);
void ParseNTP(char *Name);
void CheckCmdQueue(long Timeout);
void WiFiESP_PutQueue(char cmd);



void DispDNS_QueryBuf(void);
void DispDNSanswerBuf(void);
void DispNTP_QueryBuf(void);
void DispNTPAnswerBuf(void);




#endif

/**************************************************************************************************
  End of File
**************************************************************************************************/